@@ -12,20 +12,17 @@ not restricted to '*'.
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DefaultSignatures #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE GADTs #-}
 {-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE PolyKinds #-}
 {-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE Safe #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeApplications #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE TypeOperators #-}
-#if MIN_VERSION_base(4,9,0)
-{-# LANGUAGE Safe #-}
-#else
-{-# LANGUAGE Trustworthy #-}
-#endif
 module Data.Parameterized.Classes
   ( -- * Equality exports
     Equality.TestEquality(..)
@@ -68,15 +65,10 @@ import Data.Hashable
 import Data.Maybe (isJust)
 import Data.Proxy
 import Data.Type.Equality as Equality
+import Lens.Micro (Lens', Traversal')
 
 import Data.Parameterized.Compose ()
 
--- We define these type alias here to avoid importing Control.Lens
--- modules, as this apparently causes problems with the safe Hasekll
--- checking.
-type Lens' s a = forall f. Functor f => (a -> f a) -> s -> f s
-type Traversal' s a = forall f. Applicative f => (a -> f a) -> s -> f s
-
 ------------------------------------------------------------------------
 -- CoercibleF
 
@@ -84,7 +76,7 @@ type Traversal' s a = forall f. Applicative f => (a -> f a) -> s -> f s
 --   all the types of a family.  We generally use this to witness
 --   the fact that the type parameter to @rtp@ is a phantom type
 --   by giving an implementation in terms of Data.Coerce.coerce.
-class CoercibleF (rtp :: k -> *) where
+class CoercibleF (rtp :: k -> Type) where
   coerceF :: rtp a -> rtp b
 
 instance CoercibleF (Const x) where
@@ -101,12 +93,15 @@ instance CoercibleF (Const x) where
 -- type when they are equal. Thus this can be implemented over
 -- parameterized types that are unable to provide evidence that their
 -- type arguments are equal.
-class EqF (f :: k -> *) where
+class EqF (f :: k -> Type) where
   eqF :: f a -> f a -> Bool
 
 instance Eq a => EqF (Const a) where
   eqF (Const x) (Const y) = x == y
 
+instance EqF Proxy where
+  eqF _ _ = True
+
 ------------------------------------------------------------------------
 -- PolyEq
 
@@ -158,14 +153,40 @@ joinOrderingF GTF _ = GTF
 ------------------------------------------------------------------------
 -- OrdF
 
--- | A parameterized type that can be compared on distinct instances.
-class TestEquality ktp => OrdF (ktp :: k -> *) where
-  {-# MINIMAL compareF #-}
+-- | The `OrdF` class is a total ordering over parameterized types so
+-- that types with different parameters can be compared.
+--
+-- Instances of `OrdF` are expected to satisfy the following laws:
+--
+-- [__Transitivity__]: if @leqF x y && leqF y z@ = 'True', then @leqF x = z@ = @True@
+-- [__Reflexivity__]: @leqF x x@ = @True@
+-- [__Antisymmetry__]: if @leqF x y && leqF y x@ = 'True', then @testEquality x y@ = @Just Refl@
+--
+-- Note that the following operator interactions are expected to hold:
+--
+-- * @geqF x y@ iff @leqF y x@
+-- * @ltF  x y@ iff @leqF x y && testEquality x y = Nothing@
+-- * @gtF  x y@ iff @ltF y x@
+-- * @ltF  x y@ iff @compareF x y == LTF@
+-- * @gtF  x y@ iff @compareF x y == GTF@
+-- * @isJust (testEquality x y)@ iff @compareF x y == EQF@
+--
+-- Furthermore, when @x@ and @y@ both have type @(k tp)@, we expect:
+--
+-- * @toOrdering (compareF x y)@ equals @compare x y@ when @Ord (k tp)@ has an instance.
+-- * @isJust (testEquality x y)@ equals @x == y@ when @Eq (k tp)@ has an instance.
+--
+-- Minimal complete definition: either 'compareF' or 'leqF'.
+-- Using 'compareF' can be more efficient for complex types.
+class TestEquality ktp => OrdF (ktp :: k -> Type) where
+  {-# MINIMAL compareF | leqF #-}
 
-  -- | compareF compares two keys with different type parameters.
-  -- Instances must ensure that keys are only equal if the type
-  -- parameters are equal.
   compareF :: ktp x -> ktp y -> OrderingF x y
+  compareF x y =
+    case testEquality x y of
+      Just Refl -> EQF
+      Nothing | leqF x y -> LTF
+              | otherwise -> GTF
 
   leqF :: ktp x -> ktp y -> Bool
   leqF x y =
@@ -197,7 +218,7 @@ class TestEquality ktp => OrdF (ktp :: k -> *) where
 
 -- | Compare two values, and if they are equal compare the next values,
 -- otherwise return LTF or GTF
-lexCompareF :: forall j k (f :: j -> *) (a :: j) (b :: j) (c :: k) (d :: k)
+lexCompareF :: forall j k (f :: j -> Type) (a :: j) (b :: j) (c :: k) (d :: k)
              .  OrdF f
             => f a
             -> f b
@@ -208,7 +229,7 @@ lexCompareF x y = joinOrderingF (compareF x y)
 -- | If the \"outer\" functor has an 'OrdF' instance, then one can be generated
 -- for the \"inner\" functor. The type-level evidence of equality is deduced
 -- via generativity of @g@, e.g. the inference @g x ~ g y@ implies @x ~ y@.
-ordFCompose :: forall k l (f :: k -> *) (g :: l -> k) x y.
+ordFCompose :: forall k l (f :: k -> Type) (g :: l -> k) x y.
                 (forall w z. f w -> f z -> OrderingF w z)
             -> Compose f g x
             -> Compose f g y
@@ -230,7 +251,7 @@ instance OrdF f => OrdF (Compose f g) where
 -- To implement @'ShowF' g@, one should implement an instance @'Show'
 -- (g tp)@ for all argument types @tp@, then write an empty instance
 -- @instance 'ShowF' g@.
-class ShowF (f :: k -> *) where
+class ShowF (f :: k -> Type) where
   -- | Provides a show instance for each type.
   withShow :: p f -> q tp -> (Show (f tp) => a) -> a
 
@@ -250,11 +271,13 @@ showsF x = showsPrecF 0 x
 
 instance Show x => ShowF (Const x)
 
+instance ShowF Proxy
+
 ------------------------------------------------------------------------
 -- IxedF
 
-type family IndexF       (m :: *) :: k -> *
-type family IxValueF     (m :: *) :: k -> *
+type family IndexF       (m :: Type) :: k -> Type
+type family IxValueF     (m :: Type) :: k -> Type
 
 -- | Parameterized generalization of the lens @Ixed@ class.
 class IxedF k m where
@@ -293,7 +316,7 @@ defaultSalt = 0x087fc72c
 {-# INLINE defaultSalt #-}
 
 -- | A parameterized type that is hashable on all instances.
-class HashableF (f :: k -> *) where
+class HashableF (f :: k -> Type) where
   hashWithSaltF :: Int -> f tp -> Int
 
   -- | Hash with default salt.
@@ -323,7 +346,7 @@ instance OrdF f => Ord (TypeAp f tp) where
 instance ShowF f => Show (TypeAp f tp) where
   showsPrec p (TypeAp x) = showsPrecF p x
 
-instance HashableF f => Hashable (TypeAp f tp) where
+instance (HashableF f, TestEquality f) => Hashable (TypeAp f tp) where
   hashWithSalt s (TypeAp x) = hashWithSaltF s x
 
 ------------------------------------------------------------------------
@@ -333,5 +356,8 @@ instance HashableF f => Hashable (TypeAp f tp) where
 -- kind), a type constructor @f@ of kind @k -> *@ (typically a GADT of
 -- singleton types indexed by @k@), and an index parameter @ctx@ of
 -- kind @k@.
-class KnownRepr (f :: k -> *) (ctx :: k) where
+class KnownRepr (f :: k -> Type) (ctx :: k) where
   knownRepr :: f ctx
+
+instance KnownRepr Proxy ctx where
+  knownRepr = Proxy
