@@ -56,6 +56,7 @@ module Data.Parameterized.Context
   , Data.Parameterized.Context.last
   , Data.Parameterized.Context.view
   , Data.Parameterized.Context.take
+  , Data.Parameterized.Context.drop
   , forIndexM
   , generateSome
   , generateSomeM
@@ -63,6 +64,9 @@ module Data.Parameterized.Context
   , traverseAndCollect
   , traverseWithIndex_
   , dropPrefix
+  , unzip
+  , flattenAssignment
+  , flattenSize
 
     -- * Context extension and embedding utilities
   , CtxEmbedding(..)
@@ -75,6 +79,7 @@ module Data.Parameterized.Context
   , extendEmbeddingRight
   , extendEmbeddingBoth
   , appendEmbedding
+  , appendEmbeddingLeft
   , ctxeSize
   , ctxeAssignment
 
@@ -95,12 +100,17 @@ module Data.Parameterized.Context
   , i1of6, i2of6, i3of6, i4of6, i5of6, i6of6
   ) where
 
-import           Control.Applicative (liftA2)
-import           Control.Lens hiding (Index, (:>), Empty)
+import           Prelude hiding (unzip)
+
+import qualified Control.Applicative as App (liftA2)
+import           Data.Function ((&))
 import           Data.Functor (void)
+import           Data.Functor.Product (Product(Pair))
+import           Data.Kind
 import qualified Data.Vector as V
 import qualified Data.Vector.Mutable as MV
 import           GHC.TypeLits (Nat, type (-))
+import           Lens.Micro (Lens', Lens, lens, (^.), (%~))
 
 import           Data.Parameterized.Classes
 import           Data.Parameterized.Some
@@ -155,7 +165,6 @@ toVector a f = V.create $ do
   return vm
 {-# INLINABLE toVector #-}
 
-
 -- | Utility function for testing if @xs@ is an assignment with
 --   `prefix` as a prefix, and computing the tail of xs
 --   not in the prefix, if so.
@@ -185,7 +194,37 @@ dropPrefix xs0 prefix err = go xs0 (sizeInt (size xs0))
       Just Refl -> success Empty
       Nothing   -> err
 
+-- | Unzip an assignment of pairs into a pair of assignments.
+--
+-- This is the inverse of @'zipWith' 'Pair'@.
+unzip :: Assignment (Product f g) ctx -> (Assignment f ctx, Assignment g ctx)
+unzip fgs =
+  case viewAssign fgs of
+    AssignEmpty -> (empty, empty)
+    AssignExtend rest (Pair f g) ->
+      let (fs, gs) = unzip rest
+      in (extend fs f, extend gs g)
 
+-- | Flattens a nested assignment over a context of contexts @ctxs :: Ctx (Ctx
+-- a)@ into a flat assignment over the flattened context @CtxFlatten ctxs@.
+flattenAssignment ::
+  Assignment (Assignment f) ctxs ->
+  Assignment f (CtxFlatten ctxs)
+flattenAssignment ctxs =
+  case viewAssign ctxs of
+    AssignEmpty -> empty
+    AssignExtend ctxs' ctx -> flattenAssignment ctxs' <++> ctx
+
+-- | Given the size of each context in @ctxs@, returns the size of @CtxFlatten
+-- ctxs@.  You can obtain the former from any nested assignment @Assignment
+-- (Assignment f) ctxs@, by calling @fmapFC size@.
+flattenSize ::
+  Assignment Size ctxs ->
+  Size (CtxFlatten ctxs)
+flattenSize a =
+  case viewAssign a of
+    AssignEmpty -> zeroSize
+    AssignExtend b s -> addSize (flattenSize b) s
 
 
 --------------------------------------------------------------------------------
@@ -203,10 +242,7 @@ pattern (:>) :: () => ctx' ~ (ctx ::> tp) => Assignment f ctx -> f tp -> Assignm
 pattern (:>) a v <- (viewAssign -> AssignExtend a v)
   where a :> v = extend a v
 
--- The COMPLETE pragma was not defined until ghc 8.2.*
-#if MIN_VERSION_base(4,10,0)
 {-# COMPLETE (:>), Empty :: Assignment  #-}
-#endif
 
 --------------------------------------------------------------------------------
 -- Views
@@ -238,11 +274,16 @@ last x =
 view :: forall f ctx . Assignment f ctx -> AssignView f ctx
 view = viewAssign
 
+-- | Return the prefix of an appended 'Assignment'
 take :: forall f ctx ctx'. Size ctx -> Size ctx' -> Assignment f (ctx <+> ctx') -> Assignment f ctx
 take sz sz' asgn =
   let diff = appendDiff sz' in
   generate sz (\i -> asgn ! extendIndex' diff i)
 
+-- | Return the suffix of an appended 'Assignment'
+drop :: forall f ctx ctx'. Size ctx -> Size ctx' -> Assignment f (ctx <+> ctx') -> Assignment f ctx'
+drop sz sz' asgn = generate sz' (\i -> asgn ! extendIndexAppendLeft sz sz' i)
+
 --------------------------------------------------------------------------------
 -- Context embedding.
 
@@ -261,23 +302,23 @@ data CtxEmbedding (ctx :: Ctx k) (ctx' :: Ctx k)
 --   ExtendBoth :: CtxEmbedding ctx ctx' -> CtxEmbedding (ctx ::> tp) (ctx' ::> tp)
 --   ExtendOne  :: CtxEmbedding ctx ctx' -> CtxEmbedding ctx (ctx' ::> tp)
 
-ctxeSize :: Simple Lens (CtxEmbedding ctx ctx') (Size ctx')
+ctxeSize :: Lens' (CtxEmbedding ctx ctx') (Size ctx')
 ctxeSize = lens _ctxeSize (\s v -> s { _ctxeSize = v })
 
 ctxeAssignment :: Lens (CtxEmbedding ctx1 ctx') (CtxEmbedding ctx2 ctx')
                        (Assignment (Index ctx') ctx1) (Assignment (Index ctx') ctx2)
 ctxeAssignment = lens _ctxeAssignment (\s v -> s { _ctxeAssignment = v })
 
-class ApplyEmbedding (f :: Ctx k -> *) where
+class ApplyEmbedding (f :: Ctx k -> Type) where
   applyEmbedding :: CtxEmbedding ctx ctx' -> f ctx -> f ctx'
 
-class ApplyEmbedding' (f :: Ctx k -> k' -> *) where
+class ApplyEmbedding' (f :: Ctx k -> k' -> Type) where
   applyEmbedding' :: CtxEmbedding ctx ctx' -> f ctx v -> f ctx' v
 
-class ExtendContext (f :: Ctx k -> *) where
+class ExtendContext (f :: Ctx k -> Type) where
   extendContext :: Diff ctx ctx' -> f ctx -> f ctx'
 
-class ExtendContext' (f :: Ctx k -> k' -> *) where
+class ExtendContext' (f :: Ctx k -> k' -> Type) where
   extendContext' :: Diff ctx ctx' -> f ctx v -> f ctx' v
 
 instance ApplyEmbedding' Index where
@@ -310,11 +351,16 @@ extendEmbeddingRightDiff diff (CtxEmbedding sz' assgn) = CtxEmbedding (extSize s
 extendEmbeddingRight :: CtxEmbedding ctx ctx' -> CtxEmbedding ctx (ctx' ::> tp)
 extendEmbeddingRight = extendEmbeddingRightDiff knownDiff
 
+-- | Prove that the prefix of an appended context is embeddable in it
 appendEmbedding :: Size ctx -> Size ctx' -> CtxEmbedding ctx (ctx <+> ctx')
 appendEmbedding sz sz' = CtxEmbedding (addSize sz sz') (generate sz (extendIndex' diff))
   where
   diff = appendDiff sz'
 
+-- | Prove that the suffix of an appended context is embeddable in it
+appendEmbeddingLeft :: Size ctx -> Size ctx' -> CtxEmbedding ctx' (ctx <+> ctx')
+appendEmbeddingLeft sz sz' = CtxEmbedding (addSize sz sz') (generate sz' (extendIndexAppendLeft sz sz'))
+
 extendEmbeddingBoth :: forall ctx ctx' tp. CtxEmbedding ctx ctx' -> CtxEmbedding (ctx ::> tp) (ctx' ::> tp)
 extendEmbeddingBoth ctxe = updated & ctxeAssignment %~ flip extend (nextIndex (ctxe ^. ctxeSize))
   where
@@ -375,7 +421,7 @@ instance {-# Overlaps #-} (KnownContext xs, Idx' (n-1) xs r) =>
 -- > CurryAssignment (EmptyCtx ::> a) f x = f a -> x
 -- > CurryAssignment (EmptyCtx ::> a ::> b) f x = f a -> f b -> x
 -- > CurryAssignment (EmptyCtx ::> a ::> b ::> c) f x = f a -> f b -> f c -> x
-type family CurryAssignment (ctx :: Ctx k) (f :: k -> *) (x :: *) :: * where
+type family CurryAssignment (ctx :: Ctx k) (f :: k -> Type) (x :: Type) :: Type where
    CurryAssignment EmptyCtx    f x = x
    CurryAssignment (ctx ::> a) f x = CurryAssignment ctx f (f a -> x)
 
@@ -413,7 +459,7 @@ instance Functor (Collector m w) where
   fmap _ (Collector x) = Collector x
 instance (Applicative m, Monoid w) => Applicative (Collector m w) where
   pure _ = Collector (pure mempty)
-  Collector x <*> Collector y = Collector (liftA2 (<>) x y)
+  Collector x <*> Collector y = Collector (App.liftA2 (<>) x y)
 
 -- | Visit each of the elements in an @Assignment@ in order
 --   from left to right and collect the results using the provided @Monoid@.
