@@ -27,6 +27,7 @@
 {-# LANGUAGE RoleAnnotations #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
+{-# LANGUAGE EmptyCase #-}
 {-# LANGUAGE ExistentialQuantification #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
@@ -39,10 +40,11 @@
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE TypeInType #-}
 {-# LANGUAGE TypeOperators #-}
-{-# LANGUAGE TypeInType #-}
 {-# LANGUAGE StandaloneDeriving #-}
+#if __GLASGOW_HASKELL__ < 806
+{-# LANGUAGE TypeInType #-}
+#endif
 {-# OPTIONS_HADDOCK hide #-}
 module Data.Parameterized.Context.Safe
   ( module Data.Parameterized.Ctx
@@ -56,6 +58,7 @@ module Data.Parameterized.Context.Safe
   , addSize
   , SizeView(..)
   , viewSize
+  , sizeToNatRepr
   , KnownContext(..)
     -- * Diff
   , Diff
@@ -75,8 +78,11 @@ module Data.Parameterized.Context.Safe
   , skipIndex
   , lastIndex
   , nextIndex
+  , leftIndex
+  , rightIndex
   , extendIndex
   , extendIndex'
+  , extendIndexAppendLeft
   , forIndex
   , forIndexRange
   , intIndex
@@ -104,25 +110,22 @@ module Data.Parameterized.Context.Safe
   ) where
 
 import qualified Control.Category as Cat
-import Control.DeepSeq
-import qualified Control.Lens as Lens
-import Control.Monad.Identity (Identity(..))
-import Data.Hashable
-import Data.List (intercalate)
-import Data.Maybe (listToMaybe)
-import Data.Type.Equality
-import Prelude hiding (init, map, null, replicate, succ, zipWith)
-import Data.Kind(Type)
-
-#if !MIN_VERSION_base(4,8,0)
-import Data.Functor
-import Control.Applicative (Applicative(..))
-#endif
+import           Control.DeepSeq
+import           Control.Monad.Identity (Identity(..))
+import           Data.Hashable
+import           Data.Kind (Type)
+import           Data.List (intercalate)
+import           Data.Maybe (listToMaybe)
+import           Data.Type.Equality
+import qualified Lens.Micro as Lens
+import           Prelude hiding (init, map, null, replicate, succ, zipWith)
 
 import Data.Parameterized.Classes
 import Data.Parameterized.Ctx
+import Data.Parameterized.NatRepr
 import Data.Parameterized.Some
 import Data.Parameterized.TraversableFC
+import Data.Parameterized.TraversableFC.WithIndex
 
 ------------------------------------------------------------------------
 -- Size
@@ -169,6 +172,16 @@ viewSize :: Size ctx -> SizeView ctx
 viewSize SizeZero = ZeroSize
 viewSize (SizeSucc s) = IncSize s
 
+-- | Convert a 'Size' into a 'NatRepr'.
+sizeToNatRepr :: Size items -> NatRepr (CtxSize items)
+sizeToNatRepr sz =
+  case viewSize sz of
+    ZeroSize -> knownRepr
+    IncSize sz' ->
+      let oldRep = sizeToNatRepr sz'
+      in case plusComm (knownRepr :: NatRepr 1) oldRep of
+           Refl -> incNat oldRep
+
 ------------------------------------------------------------------------
 -- Size
 
@@ -269,6 +282,9 @@ indexVal (IndexThere idx) = indexVal idx
 instance Eq (Index ctx tp) where
   idx1 == idx2 = isJust (testEquality idx1 idx2)
 
+instance EqF (Index ctx) where
+  eqF idx1 idx2 = isJust (testEquality idx1 idx2)
+
 instance TestEquality (Index ctx) where
   testEquality (IndexHere _) (IndexHere _) = Just Refl
   testEquality (IndexHere _) (IndexThere _) = Nothing
@@ -303,17 +319,38 @@ nextIndex sz = IndexHere sz
 lastIndex :: Size (ctx ::> tp) -> Index (ctx ::> tp) tp
 lastIndex (SizeSucc s) = IndexHere s
 
+-- | Adapts an index in the left hand context of an append operation.
+leftIndex :: Size r -> Index l tp -> Index (l <+> r) tp
+leftIndex sr il = extendIndex' (appendDiff sr) il
+
+-- | Adapts an index in the right hand context of an append operation.
+rightIndex :: Size l -> Size r -> Index r tp -> Index (l <+> r) tp
+rightIndex sl sr ir =
+  case viewIndex sr ir of
+    IndexViewInit i -> skipIndex (rightIndex sl (decSize sr) i)
+    IndexViewLast s -> lastIndex (incSize (addSize sl s))
+
 {-# INLINE extendIndex #-}
 extendIndex :: KnownDiff l r => Index l tp -> Index r tp
 extendIndex = extendIndex' knownDiff
 
 {-# INLINE extendIndex' #-}
+-- | Compute an 'Index' into a context @r@ from an 'Index' into
+-- a sub-context @l@ of @r@.
 extendIndex' :: Diff l r -> Index l tp -> Index r tp
 extendIndex' DiffHere idx = idx
 extendIndex' (DiffThere diff) idx = IndexThere (extendIndex' diff idx)
 
--- | Given a size @n@, an initial value @v0@, and a function @f@, the
--- expression @forIndex n v0 f@ calls @f@ on each index less than @n@
+{-# INLINE extendIndexAppendLeft #-}
+-- | Compute an 'Index' into an appended context from an 'Index' into
+-- its suffix.
+extendIndexAppendLeft :: Size l -> Size r -> Index r tp -> Index (l <+> r) tp
+extendIndexAppendLeft sz sz' idx = case viewIndex sz' idx of
+  IndexViewLast _ -> lastIndex (addSize sz sz')
+  IndexViewInit idx' -> skipIndex (extendIndexAppendLeft sz (decSize sz') idx')
+
+-- | Given a size @n@, a function @f@, and an initial value @v0@, the
+-- expression @forIndex n f v0@ calls @f@ on each index less than @n@
 -- starting from @0@ and @v0@, with the value @v@ obtained from the
 -- last call.
 forIndex :: forall ctx r
@@ -354,9 +391,9 @@ forIndexRangeImpl _ SizeZero _ _ r = r
 forIndexRangeImpl i (SizeSucc sz) d f r =
   forIndexRangeImpl (i-1) sz (LDiffThere d) f r
 
--- | Given an index @i@, size @n@, a function @f@, value @v@, and a
--- function @f@, the expression @forIndexRange i n f v@ is equivalent
--- to @v@ when @i >= sizeInt n@, and @f i (forIndexRange (i+1) n v)@
+-- | Given an index @i@, size @n@, a function @f@, and a value @v@,
+-- the expression @forIndexRange i n f v@ is equivalent
+-- to @v@ when @i >= sizeInt n@, and @f i (forIndexRange (i+1) n f v)@
 -- otherwise.
 forIndexRange :: Int
               -> Size ctx
@@ -464,6 +501,9 @@ replicate n c = generate n (\_ -> c)
 empty :: Assignment f 'EmptyCtx
 empty = AssignmentEmpty
 
+-- n.b. see 'singleton' in Data/Parameterized/Context.hs
+
+-- | Extend an indexed vector with a new entry.
 extend :: Assignment f ctx -> f tp -> Assignment f (ctx '::> tp)
 extend asgn e = AssignmentExtend asgn e
 
@@ -481,18 +521,12 @@ adjustM f = go (\x -> x)
   go :: (forall tp'. g tp' -> f tp') -> Index ctx' tp -> Assignment g ctx' -> m (Assignment f ctx')
   go g (IndexHere _)     (AssignmentExtend asgn x) = AssignmentExtend (map g asgn) <$> f (g x)
   go g (IndexThere idx)  (AssignmentExtend asgn x) = flip AssignmentExtend (g x)   <$> go g idx asgn
-#if !MIN_VERSION_base(4,9,0)
--- GHC 7.10.3 and early does not recognize that the above definition is complete,
--- and so need the equation below.  GHC 8.0.1 does not require the additional
--- equation.
-  go _ _ _ = error "SafeTypeContext.adjustM: impossible!"
-#endif
 
 type instance IndexF   (Assignment (f :: k -> Type) ctx) = Index ctx
 type instance IxValueF (Assignment (f :: k -> Type) ctx) = f
 
 instance forall k (f :: k -> Type) ctx. IxedF k (Assignment f ctx) where
-  ixF :: Index ctx x -> Lens.Lens' (Assignment f ctx) (f x)
+  ixF :: Index ctx x -> Lens.Traversal' (Assignment f ctx) (f x)
   ixF idx f = adjustM f idx
 
 instance forall k (f :: k -> Type) ctx. IxedF' k (Assignment f ctx) where
@@ -502,7 +536,7 @@ instance forall k (f :: k -> Type) ctx. IxedF' k (Assignment f ctx) where
 idxlookup :: (forall tp. a tp -> b tp) -> Assignment a ctx -> forall tp. Index ctx tp -> b tp
 idxlookup f (AssignmentExtend _   x) (IndexHere _) = f x
 idxlookup f (AssignmentExtend ctx _) (IndexThere idx) = idxlookup f ctx idx
-idxlookup _ AssignmentEmpty _ = error "Data.Parameterized.Context.Safe.lookup: impossible case"
+idxlookup _ AssignmentEmpty idx = case idx of {}
 
 -- | Return value of assignment.
 (!) :: Assignment f ctx -> Index ctx tp -> f tp
@@ -516,6 +550,11 @@ a !^ i = a ! extendIndex i
 instance TestEquality f => Eq (Assignment f ctx) where
   x == y = isJust (testEquality x y)
 
+instance EqF f => EqF (Assignment f) where
+  eqF AssignmentEmpty AssignmentEmpty = True
+  eqF (AssignmentExtend ctx1 x1) (AssignmentExtend ctx2 x2) =
+    eqF ctx1 ctx2 && eqF x1 x2
+
 testEq :: (forall x y. f x -> f y -> Maybe (x :~: y))
        -> Assignment f cxt1 -> Assignment f cxt2 -> Maybe (cxt1 :~: cxt2)
 testEq _ AssignmentEmpty AssignmentEmpty = Just Refl
@@ -530,7 +569,7 @@ testEq _ AssignmentEmpty AssignmentExtend{} = Nothing
 testEq _ AssignmentExtend{} AssignmentEmpty = Nothing
 
 instance TestEqualityFC Assignment where
-   testEqualityFC = testEq
+   testEqualityFC f = testEq f
 instance TestEquality f => TestEquality (Assignment f) where
    testEquality x y = testEq testEquality x y
 instance TestEquality f => PolyEq (Assignment f x) (Assignment f y) where
@@ -551,7 +590,7 @@ compareAsgn test (AssignmentExtend ctx1 x) (AssignmentExtend ctx2 y) =
               EQF -> EQF
 
 instance OrdFC Assignment where
-  compareFC = compareAsgn
+  compareFC f = compareAsgn f
 
 instance OrdF f => OrdF (Assignment f) where
   compareF = compareAsgn compareF
@@ -565,10 +604,10 @@ instance Hashable (Index ctx tp) where
 instance HashableF (Index ctx) where
   hashWithSaltF s i = hashWithSalt s (indexVal i)
 
-instance HashableF f => HashableF (Assignment f) where
+instance (HashableF f, TestEquality f) => HashableF (Assignment f) where
   hashWithSaltF = hashWithSalt
 
-instance HashableF f => Hashable (Assignment f ctx) where
+instance (HashableF f, TestEquality f) => Hashable (Assignment f ctx) where
   hashWithSalt s AssignmentEmpty = s
   hashWithSalt s (AssignmentExtend asgn x) = (s `hashWithSalt` asgn) `hashWithSaltF` x
 
@@ -584,11 +623,20 @@ instance FoldableFC Assignment where
   foldMapFC = foldMapFCDefault
 
 instance TraversableFC Assignment where
-  traverseFC = traverseF
+  traverseFC f = traverseF f
+
+instance FunctorFCWithIndex Assignment where
+  imapFC = imapFCDefault
+
+instance FoldableFCWithIndex Assignment where
+  ifoldMapFC = ifoldMapFCDefault
+
+instance TraversableFCWithIndex Assignment where
+  itraverseFC = traverseWithIndex
 
 -- | Map assignment
 map :: (forall tp . f tp -> g tp) -> Assignment f c -> Assignment g c
-map = fmapFC
+map f = fmapFC f
 
 traverseF :: forall k (f:: k -> Type) (g::k -> Type) (m:: Type -> Type) (c::Ctx k)
            . Applicative m
@@ -602,7 +650,7 @@ traverseF f (AssignmentExtend asgn x) = pure AssignmentExtend <*> traverseF f as
 toList :: (forall tp . f tp -> a)
        -> Assignment f c
        -> [a]
-toList = toListFC
+toList f = toListFC f
 
 zipWithM :: Applicative m
          => (forall tp . f tp -> g tp -> m (h tp))
@@ -621,6 +669,7 @@ zipWith :: (forall x . f x -> g x -> h x)
 zipWith f = \x y -> runIdentity $ zipWithM (\u v -> pure (f u v)) x y
 {-# INLINE zipWith #-}
 
+-- | This is a specialization of 'itraverseFC'.
 traverseWithIndex :: Applicative m
                   => (forall tp . Index ctx tp -> f tp -> m (g tp))
                   -> Assignment f ctx
