@@ -1,5 +1,6 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DataKinds #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE InstanceSigs #-}
 {-# LANGUAGE GADTs #-}
@@ -11,10 +12,11 @@
 {-# LANGUAGE Trustworthy #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE TypeOperators #-}
-{-# LANGUAGE TypeInType #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE StandaloneDeriving #-}
+#if __GLASGOW_HASKELL__ < 806
 {-# LANGUAGE TypeInType #-}
+#endif
 {-# OPTIONS_HADDOCK hide #-}
 module Data.Parameterized.Context.Unsafe
   ( module Data.Parameterized.Ctx
@@ -29,6 +31,7 @@ module Data.Parameterized.Context.Unsafe
   , addSize
   , SizeView(..)
   , viewSize
+  , sizeToNatRepr
     -- * Diff
   , Diff
   , noDiff
@@ -47,8 +50,11 @@ module Data.Parameterized.Context.Unsafe
   , skipIndex
   , lastIndex
   , nextIndex
+  , leftIndex
+  , rightIndex
   , extendIndex
   , extendIndex'
+  , extendIndexAppendLeft
   , forIndex
   , forIndexRange
   , intIndex
@@ -84,21 +90,25 @@ module Data.Parameterized.Context.Unsafe
 import qualified Control.Category as Cat
 import           Control.DeepSeq
 import           Control.Exception
-import qualified Control.Lens as Lens
 import           Control.Monad.Identity (Identity(..))
 import           Data.Bits
 import           Data.Coerce
 import           Data.Hashable
+import           Data.Kind (Type)
 import           Data.List (intercalate)
 import           Data.Proxy
+import qualified Lens.Micro as Lens
 import           Unsafe.Coerce
-import           Data.Kind(Type)
 
+import           Data.Parameterized.Axiom
 import           Data.Parameterized.Classes
 import           Data.Parameterized.Ctx
 import           Data.Parameterized.Ctx.Proofs
+import           Data.Parameterized.NatRepr
+import           Data.Parameterized.NatRepr.Internal (NatRepr(NatRepr))
 import           Data.Parameterized.Some
 import           Data.Parameterized.TraversableFC
+import           Data.Parameterized.TraversableFC.WithIndex
 
 ------------------------------------------------------------------------
 -- Size
@@ -133,6 +143,10 @@ viewSize :: Size ctx -> SizeView ctx
 viewSize (Size 0) = unsafeCoerce ZeroSize
 viewSize (Size n) = assert (n > 0) (unsafeCoerce (IncSize (Size (n-1))))
 
+-- | Convert a 'Size' into a 'NatRepr'.
+sizeToNatRepr :: Size items -> NatRepr (CtxSize items)
+sizeToNatRepr (Size n) = NatRepr (fromIntegral n)
+
 instance Show (Size ctx) where
   show (Size i) = show i
 
@@ -233,9 +247,12 @@ type role Index nominal nominal
 instance Eq (Index ctx tp) where
   Index i == Index j = i == j
 
+instance EqF (Index ctx) where
+  eqF (Index i) (Index j) = i == j
+
 instance TestEquality (Index ctx) where
   testEquality (Index i) (Index j)
-    | i == j = Just (unsafeCoerce Refl)
+    | i == j = Just unsafeAxiom
     | otherwise = Nothing
 
 instance Ord (Index ctx tp) where
@@ -263,17 +280,36 @@ nextIndex n = Index (sizeInt n)
 lastIndex :: Size (ctx ::> tp) -> Index (ctx ::> tp) tp
 lastIndex n = Index (sizeInt n - 1)
 
+-- | Adapts an index in the left hand context of an append operation.
+leftIndex :: Size r -> Index l tp -> Index (l <+> r) tp
+leftIndex _ (Index il) = Index il
+
+-- | Adapts an index in the right hand context of an append operation.
+rightIndex :: Size l -> Size r -> Index r tp -> Index (l <+> r) tp
+rightIndex (Size sl) _ (Index ir) = Index (sl + ir)
+
 {-# INLINE extendIndex #-}
 extendIndex :: KnownDiff l r => Index l tp -> Index r tp
 extendIndex = extendIndex' knownDiff
 
 {-# INLINE extendIndex' #-}
+-- | Compute an 'Index' into a context @r@ from an 'Index' into
+-- a sub-context @l@ of @r@.
 extendIndex' :: Diff l r -> Index l tp -> Index r tp
 extendIndex' _ = unsafeCoerce
 
--- | Given a size @n@, an initial value @v0@, and a function @f@, the
--- expression @forIndex n v0 f@ is equivalent to @v0@ when @n@ is
--- zero, and @f (forIndex (n-1) v0) (n-1)@ otherwise.
+{-# INLINE extendIndexAppendLeft #-}
+-- | Compute an 'Index' into an appended context from an 'Index' into
+-- its suffix.
+extendIndexAppendLeft :: Size l -> Size r -> Index r tp -> Index (l <+> r) tp
+extendIndexAppendLeft (Size l) _ (Index idx) = Index (idx + l)
+
+-- | Given a size @n@, a function @f@, and an initial value @v0@, the
+-- expression @forIndex n f v0@ is equivalent to @v0@ when @n@ is
+-- zero, and @f (forIndex (n-1) f v0) n@ otherwise.  Unlike the safe
+-- version, which starts from 'Index' @0@ and increments 'Index'
+-- values, this version starts at 'Index' @(n-1)@ and decrements
+-- 'Index' values to 'Index' @0@.
 forIndex :: forall ctx r
           . Size ctx
          -> (forall tp . r -> Index ctx tp -> r)
@@ -284,9 +320,9 @@ forIndex n f r =
     ZeroSize -> r
     IncSize p -> f (forIndex p (coerce f) r) (nextIndex p)
 
--- | Given an index @i@, size @n@, a function @f@, value @v@, and a
--- function @f@, the expression @forIndex i n f v@ is equivalent to
--- @v@ when @i >= sizeInt n@, and @f i (forIndexRange (i+1) n v)@
+-- | Given an index @i@, size @n@, a function @f@, and a value @v@,
+-- the expression @forIndex i n f v@ is equivalent to
+-- @v@ when @i >= sizeInt n@, and @f i (forIndexRange (i+1) n f v)@
 -- otherwise.
 forIndexRange :: forall ctx r
                . Int
@@ -385,17 +421,10 @@ instance TestEqualityFC (BalancedTree h) where
     Refl <- testEqualityFC test x1 y1
     Refl <- testEqualityFC test x2 y2
     return Refl
-#if !MIN_VERSION_base(4,9,0)
-  testEqualityFC _ _ _ = Nothing
-#endif
 
 instance OrdFC (BalancedTree h) where
   compareFC test (BalLeaf x) (BalLeaf y) =
     joinOrderingF (test x y) $ EQF
-#if !MIN_VERSION_base(4,9,0)
-  compareFC _ BalLeaf{} _ = LTF
-  compareFC _ _ BalLeaf{} = GTF
-#endif
   compareFC test (BalPair x1 x2) (BalPair y1 y2) =
     joinOrderingF (compareFC test x1 y1) $
     joinOrderingF (compareFC test x2 y2) $
@@ -509,9 +538,6 @@ bal_zipWithM f (BalLeaf x) (BalLeaf y) = BalLeaf <$> f x y
 bal_zipWithM f (BalPair x1 x2) (BalPair y1 y2) =
   BalPair <$> bal_zipWithM f x1 (unsafeCoerce y1)
           <*> bal_zipWithM f x2 (unsafeCoerce y2)
-#if !MIN_VERSION_base(4,9,0)
-bal_zipWithM _ _ _ = error "illegal args to bal_zipWithM"
-#endif
 {-# INLINABLE bal_zipWithM #-}
 
 ------------------------------------------------------------------------
@@ -782,6 +808,9 @@ generateM n f = Assignment <$> unsafe_bin_generateM (sizeInt n) 0 f
 empty :: Assignment f EmptyCtx
 empty = Assignment Empty
 
+-- n.b. see 'singleton' in Data/Parameterized/Context.hs
+
+-- | Extend an indexed vector with a new entry.
 extend :: Assignment f ctx -> f x -> Assignment f (ctx ::> x)
 extend (Assignment x) y = Assignment $ append x (BalLeaf y)
 
@@ -804,6 +833,14 @@ instance TestEqualityFC Assignment where
      Refl <- testEqualityFC test x y
      return Refl
 
+instance EqF f => EqF (Assignment f) where
+  eqF x y =
+    case (viewAssign x, viewAssign y) of
+      (AssignEmpty, AssignEmpty) ->
+        True
+      (AssignExtend ctx1 x1, AssignExtend ctx2 x2) ->
+        eqF ctx1 ctx2 && eqF x1 x2
+
 instance TestEquality f => TestEquality (Assignment f) where
   testEquality = testEqualityFC testEquality
 
@@ -826,10 +863,10 @@ instance HashableF (Index ctx) where
 instance Hashable (Index ctx tp) where
   hashWithSalt = hashWithSaltF
 
-instance HashableF f => Hashable (Assignment f ctx) where
+instance (HashableF f, TestEquality f) => Hashable (Assignment f ctx) where
   hashWithSalt s (Assignment a) = hashWithSaltF s a
 
-instance HashableF f => HashableF (Assignment f) where
+instance (HashableF f, TestEquality f) => HashableF (Assignment f) where
   hashWithSaltF = hashWithSalt
 
 instance ShowF f => Show (Assignment f ctx) where
@@ -858,7 +895,7 @@ instance forall k (f :: k -> Type) ctx. IxedF' k (Assignment (f :: k -> Type) ct
   ixF' idx f = adjustM f idx
 
 instance forall k (f :: k -> Type) ctx. IxedF k (Assignment f ctx) where
-  ixF = ixF'
+  ixF idx = ixF' idx
 
 -- This is an unsafe version of update that changes the type of the expression.
 unsafeUpdate :: Int -> Assignment f ctx -> f u -> Assignment f ctx'
@@ -905,6 +942,16 @@ instance TraversableFC Assignment where
   traverseFC = \f (Assignment x) -> Assignment <$> traverse_bin f x
   {-# INLINE traverseFC #-}
 
+instance FunctorFCWithIndex Assignment where
+  imapFC = imapFCDefault
+
+instance FoldableFCWithIndex Assignment where
+  ifoldMapFC = ifoldMapFCDefault
+
+instance TraversableFCWithIndex Assignment where
+  itraverseFC = traverseWithIndex
+
+
 traverseWithIndex :: Applicative m
                   => (forall tp . Index ctx tp -> f tp -> m (g tp))
                   -> Assignment f ctx
