@@ -129,15 +129,20 @@ use the 'Data.Parameterized.List.List' type for this purpose.
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE TypeOperators #-}
 module Data.Parameterized.List
   ( List(..)
+  , fromSomeList
+  , fromListWith
+  , fromListWithM
   , Index(..)
   , indexValue
   , (!!)
   , update
   , indexed
   , imap
+  , ifoldlM
   , ifoldr
   , izipWith
   , itraverse
@@ -148,14 +153,19 @@ module Data.Parameterized.List
   , index3
   ) where
 
-import qualified Control.Lens as Lens
+import qualified Lens.Micro as Lens
+import           Data.Foldable
+import           Data.Kind
 import           Prelude hiding ((!!))
+import           Unsafe.Coerce (unsafeCoerce)
 
 import           Data.Parameterized.Classes
+import           Data.Parameterized.Some
 import           Data.Parameterized.TraversableFC
+import           Data.Parameterized.TraversableFC.WithIndex
 
 -- | Parameterized list of elements.
-data List :: (k -> *) -> [k] -> * where
+data List :: (k -> Type) -> [k] -> Type where
   Nil  :: List f '[]
   (:<) :: f tp -> List f tps -> List f (tp : tps)
 
@@ -184,6 +194,18 @@ instance TraversableFC List where
   traverseFC _ Nil = pure Nil
   traverseFC f (h :< r) = (:<) <$> f h <*> traverseFC f r
 
+type instance IndexF   (List (f :: k -> Type) sh) = Index sh
+type instance IxValueF (List (f :: k -> Type) sh) = f
+
+instance FunctorFCWithIndex List where
+  imapFC = imap
+
+instance FoldableFCWithIndex List where
+  ifoldrFC = ifoldr
+
+instance TraversableFCWithIndex List where
+  itraverseFC = itraverse
+
 instance TestEquality f => TestEquality (List f) where
   testEquality Nil Nil = Just Refl
   testEquality (xh :< xl) (yh :< yl) = do
@@ -192,6 +214,12 @@ instance TestEquality f => TestEquality (List f) where
     pure Refl
   testEquality _ _ = Nothing
 
+instance EqF f => EqF (List f) where
+  eqF Nil Nil =
+    True
+  eqF (xh :< xl) (yh :< yl) =
+    eqF xh yh && eqF xl yl
+
 instance OrdF f => OrdF (List f) where
   compareF Nil Nil = EQF
   compareF Nil _ = LTF
@@ -201,21 +229,42 @@ instance OrdF f => OrdF (List f) where
     lexCompareF xl yl $
     EQF
 
-
 instance KnownRepr (List f) '[] where
   knownRepr = Nil
 
 instance (KnownRepr f s, KnownRepr (List f) sh) => KnownRepr (List f) (s ': sh) where
   knownRepr = knownRepr :< knownRepr
 
+-- | Apply function to list to yield a parameterized list.
+fromListWith :: forall a f . (a -> Some f) -> [a] -> Some (List f)
+fromListWith f = foldr g (Some Nil)
+  where g :: a -> Some (List f) -> Some (List f)
+        g x (Some r) = viewSome (\h -> Some (h :< r)) (f x)
+
+-- | Apply monadic action to list to yield a parameterized list.
+fromListWithM :: forall a f m
+                  .  Monad m
+                  => (a -> m (Some f))
+                  -> [a]
+                  -> m (Some (List f))
+fromListWithM f = foldrM g (Some Nil)
+  where g :: a -> Some (List f) -> m (Some (List f))
+        g x (Some r) = viewSome (\h -> Some (h :< r)) <$> f x
+
+-- | Map from list of Some to Some list
+fromSomeList :: [Some f] -> Some (List f)
+fromSomeList = fromListWith id
+
+{-# INLINABLE fromListWith #-}
+{-# INLINABLE fromListWithM #-}
+
 --------------------------------------------------------------------------------
 -- * Indexed operations
 
-
 -- | Represents an index into a type-level list. Used in place of integers to
 --   1. ensure that the given index *does* exist in the list
 --   2. guarantee that it has the given kind
-data Index :: [k] -> k -> *  where
+data Index :: [k] -> k -> Type  where
   IndexHere :: Index (x:r) x
   IndexThere :: !(Index r y) -> Index (x:r) y
 
@@ -293,6 +342,8 @@ indexed (IndexThere i) f (x :< rest) = (x :<) <$> indexed i f rest
 
 -- | Map over the elements in the list, and provide the index into
 -- each element along with the element itself.
+--
+-- This is a specialization of 'imapFC'.
 imap :: forall f g l
      . (forall x . Index l x -> f x -> g x)
      -> List f l
@@ -308,7 +359,29 @@ imap f = go id
         Nil -> Nil
         e :< rest -> f (g IndexHere) e :< go (g . IndexThere) rest
 
+-- | Left fold with an additional index.
+ifoldlM :: forall sh a b m
+        . Monad m
+        => (forall tp . b -> Index sh tp -> a tp -> m b)
+        -> b
+        -> List a sh
+        -> m b
+ifoldlM _ b Nil = pure b
+ifoldlM f b0 (a0 :< r0) = f b0 IndexHere a0 >>= go IndexHere r0
+  where
+    go :: forall tps tp
+        . Index sh tp
+       -> List a tps
+       -> b
+       -> m b
+    go _ Nil b = pure b
+    go idx (a :< rest) b =
+      let idx' = unsafeCoerce (IndexThere idx)
+       in f b idx' a >>= go idx' rest
+
 -- | Right-fold with an additional index.
+--
+-- This is a specialization of 'ifoldrFC'.
 ifoldr :: forall sh a b . (forall tp . Index sh tp -> a tp -> b -> b) -> b -> List a sh -> b
 ifoldr f seed0 l = go id l seed0
   where
@@ -341,6 +414,8 @@ izipWith f = go id
           f (g IndexHere) a b :< go (g . IndexThere) as' bs'
 
 -- | Traverse with an additional index.
+--
+-- This is a specialization of 'itraverseFC'.
 itraverse :: forall a b sh t
           . Applicative t
           => (forall tp . Index sh tp -> a tp -> t (b tp))
