@@ -19,7 +19,9 @@ Some code was adapted from containers.
 {-# LANGUAGE TypeOperators #-}
 {-# LANGUAGE Trustworthy #-}
 {-# LANGUAGE TypeFamilies #-}
+#if __GLASGOW_HASKELL__ < 806
 {-# LANGUAGE TypeInType #-}
+#endif
 module Data.Parameterized.Map
   ( MapF
     -- * Construction
@@ -80,11 +82,13 @@ module Data.Parameterized.Map
   ) where
 
 import           Control.Applicative hiding (empty)
-import           Control.Lens (Traversal', Lens')
-import           Control.Monad.Identity
+import           Control.Monad.Identity (Identity(..))
+import           Control.Monad (foldM)
 import           Data.Kind (Type)
-import           Data.List (intercalate, foldl')
+import qualified Data.List as List
 import           Data.Monoid
+import           Prelude hiding (filter, lookup, map, traverse, null)
+import           Lens.Micro (Traversal', Lens')
 
 import           Data.Parameterized.Classes
 import           Data.Parameterized.Some
@@ -104,12 +108,6 @@ import           Data.Parameterized.Utils.BinTree
   )
 import qualified Data.Parameterized.Utils.BinTree as Bin
 
-#if MIN_VERSION_base(4,8,0)
-import           Prelude hiding (filter, lookup, map, traverse, null)
-#else
-import           Prelude hiding (filter, lookup, map, null)
-#endif
-
 ------------------------------------------------------------------------
 -- * Pair
 
@@ -372,7 +370,7 @@ showMap :: (forall tp . ktp tp -> String)
         -> (forall tp . rtp tp -> String)
         -> MapF ktp rtp
         -> String
-showMap ppk ppv m = "{ " ++ intercalate ", " l ++ " }"
+showMap ppk ppv m = "{ " ++ List.intercalate ", " l ++ " }"
   where l = foldrWithKey (\k a l0 -> (ppk k ++ " -> " ++ ppv a) : l0) [] m
 
 ------------------------------------------------------------------------
@@ -524,7 +522,7 @@ updateAtKey k onNotFound onFound t = ins <$> atKey' k onNotFound onFound t
 
 -- | Create a Map from a list of pairs.
 fromList :: OrdF k => [Pair k a] -> MapF k a
-fromList = foldl' (\m (Pair k a) -> insert k a m) Data.Parameterized.Map.empty
+fromList = List.foldl' (\m (Pair k a) -> insert k a m) Data.Parameterized.Map.empty
 
 -- | Return list of key-values pairs in map in ascending order.
 toAscList :: MapF k a -> [Pair k a]
