@@ -17,14 +17,17 @@ module Data.Parameterized.Some
   , mapSome
   , traverseSome
   , traverseSome_
+  , someLens
   ) where
 
 import Data.Hashable
+import Data.Kind (Type)
 import Data.Parameterized.Classes
 import Data.Parameterized.TraversableF
+import Lens.Micro (Lens', lens, (&), (^.), (.~))
 
 
-data Some (f:: k -> *) = forall x . Some (f x)
+data Some (f:: k -> Type) = forall x . Some (f x)
 
 instance TestEquality f => Eq (Some f) where
   Some x == Some y = isJust (testEquality x y)
@@ -32,7 +35,7 @@ instance TestEquality f => Eq (Some f) where
 instance OrdF f => Ord (Some f) where
   compare (Some x) (Some y) = toOrdering (compareF x y)
 
-instance HashableF f => Hashable (Some f) where
+instance (HashableF f, TestEquality f) => Hashable (Some f) where
   hashWithSalt s (Some x) = hashWithSaltF s x
   hash (Some x) = hashF x
 
@@ -63,3 +66,8 @@ traverseSome_ f (Some x) = (\_ -> ()) `fmap` f x
 instance FunctorF     Some where fmapF     = mapSome
 instance FoldableF    Some where foldMapF  = foldMapFDefault
 instance TraversableF Some where traverseF = traverseSome
+
+-- | A lens that is polymorphic in the index may be used on a value with an
+-- existentially-quantified index.
+someLens :: (forall tp. Lens' (f tp) a) -> Lens' (Some f) a
+someLens l = lens (\(Some x) -> x ^. l) (\(Some x) v -> Some (x & l .~ v))
